//+------------------------------------------------------------------+
//|                                                        ZoPNG.mq5 |
//|                                           Copyright 2015, Zorro. |
//|                              https://www.mql5.com/ru/users/zorro |
//+------------------------------------------------------------------+
#property copyright "Copyright 2015, Zorro"
#property link      "https://www.mql5.com/ru/users/zorro"
#property version   "1.00"
#property library
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
template<typename T>
bool ZoArray2Struct(T &s,const uchar &data[],int start)
  {
   struct _ { uchar data[sizeof(s)]; } dst;
//---
   if(ArrayCopy(dst.data,data,0,start,sizeof(T))!=sizeof(T))
      return(false);
//---
   s=(T)dst;
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uint ZoCRCUpdate(uint crc,uchar &array[],uint start,uint count)
  {
   struct SCRCTable
     {      
      uint              data[256];
                        SCRCTable()
        {
         uint c;

         for(uint n=0; n<256; n++)
           {
            c=n;            
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            c=((c&1)!=0)?0xedb88320^(c>>1):c>>1;
            //---     
            data[n]=c;
           }
        }
      uint              operator[](int n) { return(data[n]); }
     };
//---   
   static SCRCTable table;
//---
   if(ArraySize(array)<int(start+count))
      return(0);
//---
   for(uint n=0; n<count; n++,start++)
       crc=table[(crc^array[start])&0xff]^(crc>>8);
//---
   return crc;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
uint ZoCRC(uchar &array[],uint start,uint count)
  {
   return(ZoCRCUpdate(0xFFFFFFFF,array,start,count)^0xFFFFFFFF);
  }
//+------------------------------------------------------------------+
//|                                                             |
//+------------------------------------------------------------------+
struct ZoStream
  {
   uchar             data[];
   uint              size;
   uint              pos;

   bool              IsEmpty(void) const { return(size==0);   }
   bool              IsEOF  (void) const { return(pos>=size); }
   bool              IsValid(void) const { return((uint)ArraySize(data)>=size && pos<size); }
   uint              Avail  (void) const { return(IsEOF()?0:size-pos); }

   bool              FromFile(const string path,bool common=false);
   bool              ReadArray(uchar &array[],int count)
     {
      if(!IsValid() || (pos+count)>size)
         return(false);
      //---
      int read=ArrayCopy(array,data,0,pos,count);
      pos+=read;
      return(read==count);
     }
   long              ReadLong(void)
     {
      long res=0;
      //---
      if(IsValid())
         for(int i=0;i<sizeof(long) && pos<size;i++,pos++)
            res=(res<<8)+data[pos];
      //---
      return(res);
     }
   int               ReadInt(void)
     {
      int res=0;
      //---
      if(IsValid())
         for(int i=0;i<sizeof(int) && pos<size;i++,pos++)
            res=(res<<8)+data[pos];
      //---
      return(res);
     }

   template<typename T>
   T operator>>(T)
     {
      T res=0;
      //---
      if(IsValid())
         for(int i=0;i<sizeof(T) && pos<size;i++,pos++)
            res=T((res<<8)+data[pos]);
      //---
      return(res);
     }
  };

#define INT  (int(0))
#define LONG (long(0))
#define WORD (short(0))
#define BYTE (char(0))
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum ZoPNGLoadError
  {
   ZoPNGError_NoError         =0,            // Success
   ZoPNGError_Unknown         =0x80000000,   // Unknonwn error
   ZoPNGError_BadStream       =0x80000001,   // Stream is invalid
   ZoPNGError_Unsupported     =0x80000002,   // Unsupported PNG format
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
struct ZoPNGInfo
  {
   uint              width;
   uint              height;
   uchar             bit_depth;
   uchar             clr_type;
   uchar             compression;   
   uchar             filter;
   uchar             interlace;
   int               bpp;        // bytes per pixel
   int               bpr;        // bytes per row
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ZoPNGSingnatureCheckup(ZoStream &stream)
  {
//---  PNG
   const long sign=((((((((((((long(137<<8)+80)<<8)+78)<<8)+71)<<8)+13)<<8)+10)<<8)+26)<<8)+10;
//---  
   return(stream>>LONG==sign);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ZoPNGNextChunk(ZoStream &stream,uint &chunk_id,uchar &png_chunk[])
  {
   uint length=uint(stream>>INT);
//---  
   if(!stream.IsValid() || (length+stream.pos+sizeof(uint)+sizeof(ushort))>stream.size)
      return(false);
//---  CRC
   uint crc=ZoCRC(stream.data,stream.pos,length+sizeof(int));
//---  ID 
   chunk_id=uint(stream>>INT);
//---  
   if(!stream.ReadArray(png_chunk,length))
      return(false);
//---  CRC
   return(uint(stream>>INT)==crc);
  }
//+------------------------------------------------------------------+
//|   PNG                                       |
//+------------------------------------------------------------------+
bool ZoPNGLoad(uchar &png[],uint start,uint count,uint &image[],int &width,int &height,ZoPNGLoadError &error) export
  {
   ZoStream stream;
   ArrayCopy(stream.data,png,0,start,count);
   stream.pos =0;
   stream.size=count;
//---
   error=ZoPNGError_Unknown;
//---    
   if(!ZoPNGSingnatureCheckup(stream))
     {
      error=ZoPNGError_BadStream;
      return(false);
     }
//---
   ZoPNGInfo   header={};
   bool        is_header=false;
   uint        length,crc,id;
   uchar       data[],prev[];
   uint        pal[];
//---
   for(int chunk=0;chunk>=0;chunk++)
     {
      //---   
      length=uint(stream>>INT);
      //---   
      if(stream.Avail()<(length+sizeof(id)+sizeof(crc)))
        {
         error=ZoPNGError_BadStream;
         return(false);
        }
      //---  CRC
      crc=ZoCRC(stream.data,stream.pos,length+sizeof(id));
      //---  ID
      id=uint(stream>>INT);
      //---  
      switch(id)
        {
         // tRNS
         case 0x74524e53:
            //---     
            if(!is_header)
              {
               error=ZoPNGError_BadStream;
               return(false);
              }
            //---
            for(int i=0,c=ArraySize(pal);i<c;i++)
               pal[i]=(pal[i]&0xFFFFFF)|(stream.data[stream.pos+i]<<24);
            //---
            stream.pos+=length;
            break;

         // PLTE
         case 0x504c5445:
            //---     
            if(!is_header)
              {
               error=ZoPNGError_BadStream;
               return(false);
              }
            //---
            ArrayResize(pal,length/3);
            //---
            for(uint i=0,c=length/3;i<c;i++)
               pal[i]=0xFF000000+(stream.data[stream.pos+i*3+0]<<16)+(stream.data[stream.pos+i*3+1]<<8)+stream.data[stream.pos+i*3+2];
            //---
            stream.pos+=length;
            break;

         // IEND
         case 0x49454e44:
            //---     
            if(!is_header)
              {
               error=ZoPNGError_BadStream;
               return(false);
              }
            //---
            {
               uchar result[];
               //uchar key[]={1,0,0,0};
               int res=CryptDecode(CRYPT_ARCH_ZIP,data,result,result);
               
               if(res==0)
                  return(false);
               //---
               width =int(header.width);
               height=int(header.height);
               //---
               ArrayResize(image,width*height);
               ArrayResize(prev,width);
               ArrayInitialize(prev,0);
               //---
               for(int y=0,offset=0;y<height;y++)
                 {
                  uchar filter=result[offset++];
                  
                  int byte_width=0;
                  switch(header.clr_type)
                    {
                     case 6: byte_width=1;
                     case 2: byte_width+=3; break;
                     case 3: byte_width=1;  break;
                    }
                  //---
                  switch(filter)
                    {
                     case 1:
                        for(int i=0;i<(width-1)*byte_width;i++)
                           result[offset+i+byte_width]+=result[offset+i];
                        break; // sub
                     case 2:
                        for(int i=0;i<width*byte_width;i++)
                           result[offset+i]+=prev[i];
                        break; // up
                     case 3:
                        for(int i=0;i<byte_width;i++)
                           result[offset+i]=uchar((int)result[offset+i]+(int)prev[i]/2);
                        for(int i=byte_width;i<width*byte_width;i++)
                           result[offset+i]=uchar((int)result[offset+i]+((int)result[offset+i-byte_width]+prev[i])/2);
                        break; // avg
                        // paeth
                     case 4:
                        for(int i=0;i<byte_width;i++)
                           result[offset+i]+=prev[i];                           
                        for(int i=byte_width;i<width*byte_width;i++)
                          {
                           int a, b, c, pa, pb, pc, p;                           
                           a = result[offset+i-byte_width];
                           c = prev[i-byte_width];
                           b = prev[i];                                    
                           p = b - c;
                           pc = a - c;                     
                           pa = p < 0 ? -p : p;
                           pb = pc < 0 ? -pc : pc;
                           pc = (p + pc) < 0 ? -(p + pc) : p + pc;                     
                           if (pb < pa) { pa = pb; a = b; }
                           if (pc < pa) a = c;
                           result[offset+i]=uchar(result[offset+i]+a);
                          }
                        break;
                     default:
                        error=ZoPNGError_Unsupported;
                        return(false);
                    }
                    
                  ArrayCopy(prev,result,0,offset,width*byte_width);
                  //---
                  for(int x=0;x<width;x++)
                    {
                     uint clr;
                     //---
                     switch(header.clr_type)
                       {
                        case 6:
                           clr=(result[offset++]<<16)
                               +(result[offset++]<< 8)
                               +(result[offset++]<< 0)
                               +(result[offset++]<<24);
                           break;
                        case 2:
                           clr=(result[offset++]<<16)
                               +(result[offset++]<< 8)
                               +(result[offset++]<< 0)
                               +0xFF000000;
                           break;
                        case 3:
                           clr=pal[result[offset]];
                           break;
                        default:
                           error=ZoPNGError_Unsupported;
                           return(false);                        
                       }
                    
                     image[y*width+x]=clr;
                    }                                      
                 }
                 
               int file=FileOpen("1.bin",FILE_SHARE_READ|FILE_WRITE|FILE_BIN);
               FileWriteArray(file,image);
               FileClose(file);                 
            }
            //---
            stream.pos+=length;
            chunk=-2;
            break;
         // IHDR
         case 0x49484452:
            //--- ,   
            is_header=true;
            //---  
            header.width      =stream>>INT;
            header.height     =stream>>INT;
            header.bit_depth  =stream>>BYTE;
            header.clr_type   =stream>>BYTE;
            header.compression=stream>>BYTE;   
            header.filter     =stream>>BYTE;
            header.interlace  =stream>>BYTE;            
            //---    ,       
            if(chunk!=0 || header.width==0 || header.height==0)
              {
               error=ZoPNGError_BadStream;
               return(false);               
              }
            //---
            if(header.bit_depth!=8 || (header.clr_type!=2 && header.clr_type!=6) || header.filter!=0 || header.compression!=0 || header.interlace!=0)
              {
               error=ZoPNGError_Unsupported;
               return(false);
              }
            //---
            break;
         // IDAT
         case 0x49444154:
            ArrayCopy(data,stream.data,ArraySize(data),stream.pos,length);
            stream.pos+=length;
            break;
         
         default:
            //PrintFormat("%c%c%c%c: %x %u",(id>>24)&255,(id>>16)&255,(id>>8)&255,(id>>0)&255,id,length);
            stream.pos+=length;
            break;
        }
      //---  CRC
      if(crc!=stream>>INT)
        {
         error=ZoPNGError_BadStream;
         return(false);        
        }
     }
//---
   return(true);
  }
//+------------------------------------------------------------------+
